from typing import Sequence, Mapping, Optional, List

import attrs
from numpy.random import Generator as RandomGenerator

from vkit.utility import rng_choice
from vkit.engine.interface import Engine, EngineExecutorFactory
from .type import CharSamplerEngineInitResource, CharSamplerEngineRunConfig


@attrs.define
class CharSamplerLexiconEngineInitConfig:
    tag_to_weight: Optional[Mapping[str, float]] = None
    space_prob: float = 0.0


CharSamplerLexiconEngineInitResource = CharSamplerEngineInitResource


class CharSamplerLexiconEngine(
    Engine[
        CharSamplerLexiconEngineInitConfig,
        CharSamplerLexiconEngineInitResource,
        CharSamplerEngineRunConfig,
        Sequence[str],
    ]
):  # yapf: disable

    KEY_SPACE = '__space'

    @classmethod
    def get_type_name(cls) -> str:
        return 'lexicon'

    def __init__(
        self,
        init_config: CharSamplerLexiconEngineInitConfig,
        init_resource: Optional[CharSamplerLexiconEngineInitResource] = None,
    ):
        super().__init__(init_config, init_resource)

        assert init_resource
        self.lexicon_collection = init_resource.lexicon_collection

        weights = []
        if init_config.tag_to_weight:
            for tag in self.lexicon_collection.tags:
                assert tag in init_config.tag_to_weight
                weights.append(init_config.tag_to_weight[tag])
        else:
            for tag in self.lexicon_collection.tags:
                weights.append(len(self.lexicon_collection.tag_to_lexicons[tag]))

        self.tags_no_space = list(self.lexicon_collection.tags)
        total = sum(weights)
        self.probs_no_space = [val / total for val in weights]

        self.tags = list(self.tags_no_space)
        if init_config.space_prob > 0:
            space_weight = total * init_config.space_prob / (1 - init_config.space_prob)
            self.tags.append(self.KEY_SPACE)
            weights.append(space_weight)
            total += space_weight
        self.probs = [val / total for val in weights]

    def run(self, run_config: CharSamplerEngineRunConfig, rng: RandomGenerator) -> Sequence[str]:
        num_chars = run_config.num_chars

        if run_config.enable_aggregator_mode:
            num_chars = int(rng.integers(1, run_config.num_chars + 1))

        chars: List[str] = []
        for char_idx in range(num_chars):
            tag = rng_choice(rng, self.tags, probs=self.probs)
            if tag == self.KEY_SPACE and (char_idx == 0 or char_idx == num_chars - 1):
                tag = rng_choice(rng, self.tags_no_space, probs=self.probs_no_space)

            if tag == self.KEY_SPACE:
                chars.append(' ')
            else:
                lexicon = rng_choice(rng, self.lexicon_collection.tag_to_lexicons[tag])
                char = rng_choice(rng, lexicon.char_and_aliases)
                chars.append(char)

        return chars


char_sampler_lexicon_engine_executor_factory = EngineExecutorFactory(CharSamplerLexiconEngine)
